from functools import lru_cache

from django.http import Http404
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.text import capfirst
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from wagtail.admin import messages
from wagtail.admin.panels import (
    ObjectList,
    extract_panel_definitions_from_model_class,
)
from wagtail.admin.views import generic
from wagtail.models import Site
from wagtail.permission_policies import ModelPermissionPolicy

from .forms import SiteSwitchForm
from .models import BaseGenericSetting, BaseSiteSetting
from .registry import registry


def get_model_from_url_params(app_name, model_name):
    """
    retrieve a content type from an app_name / model_name combo.
    Throw Http404 if not a valid setting type
    """
    model = registry.get_by_natural_key(app_name, model_name)
    if model is None:
        raise Http404
    return model


@lru_cache(maxsize=None)
def get_setting_edit_handler(model):
    if hasattr(model, "edit_handler"):
        edit_handler = model.edit_handler
    else:
        if issubclass(model, BaseSiteSetting):
            panels = extract_panel_definitions_from_model_class(model, ["site"])
        elif issubclass(model, BaseGenericSetting):
            panels = extract_panel_definitions_from_model_class(model)
        else:
            raise NotImplementedError

        edit_handler = ObjectList(panels)
    return edit_handler.bind_to_model(model)


def redirect_to_relevant_instance(request, app_name, model_name):
    model = get_model_from_url_params(app_name, model_name)

    if issubclass(model, BaseSiteSetting):
        # Redirect the user to the edit page for the current site
        # (or the current request does not correspond to a site, the first site in the list)
        site_request = Site.find_for_request(request)
        site = site_request or Site.objects.first()
        if not site:
            messages.error(
                request,
                _("This setting could not be opened because there is no site defined."),
            )
            return redirect("wagtailadmin_home")
        return redirect(
            "wagtailsettings:edit",
            app_name,
            model_name,
            site.pk,
        )
    elif issubclass(model, BaseGenericSetting):
        return redirect(
            "wagtailsettings:edit",
            app_name,
            model_name,
            model.load(request_or_site=request).id,
        )
    else:
        raise NotImplementedError


class EditView(generic.EditView):
    template_name = "wagtailsettings/edit.html"
    edit_url_name = "wagtailsettings:edit"
    error_message = gettext_lazy("The setting could not be saved due to errors.")
    permission_required = "change"

    def setup(self, request, app_name, model_name, *args, **kwargs):
        self.app_name = app_name
        self.model_name = model_name
        self.model = get_model_from_url_params(app_name, model_name)
        self.permission_policy = ModelPermissionPolicy(self.model)
        self.pk = kwargs.get(self.pk_url_kwarg)
        super().setup(request, app_name, model_name, *args, **kwargs)

    def get_header_icon(self):
        return registry._model_icons.get(self.model)

    def get_object(self, queryset=None):
        self.site = None
        if issubclass(self.model, BaseSiteSetting):
            self.site = get_object_or_404(Site, pk=self.pk)
            return self.model.for_site(self.site)
        else:
            return get_object_or_404(self.model, pk=self.pk)

    def get_panel(self):
        return get_setting_edit_handler(self.model)

    def get_edit_url(self):
        return reverse(
            self.edit_url_name,
            args=(self.app_name, self.model_name, self.pk),
        )

    def get_success_buttons(self):
        return None

    def get_page_subtitle(self):
        return capfirst(self.model._meta.verbose_name)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        site_switcher = None
        if self.site and Site.objects.count() > 1:
            site_switcher = SiteSwitchForm(self.site, self.model)
            context["media"] += site_switcher.media

        context["site_switcher"] = site_switcher
        return context

    def get_success_url(self):
        return self.get_edit_url()

    def get_success_message(self):
        return capfirst(
            _("%(setting_type)s updated.")
            % {"setting_type": self.model._meta.verbose_name}
        )



from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from wagtail.admin.views.pages.bulk_actions.page_bulk_action import PageBulkAction


class DeleteBulkAction(PageBulkAction):
    display_name = _("Delete")
    action_type = "delete"
    aria_label = _("Delete selected pages")
    template_name = "wagtailadmin/pages/bulk_actions/confirm_bulk_delete.html"
    action_priority = 30
    classes = {"serious"}

    def check_perm(self, page):
        return page.permissions_for_user(self.request.user).can_delete()

    def object_context(self, page):
        return {
            "item": page,
            "descendant_count": page.get_descendant_count(),
        }

    @classmethod
    def execute_action(cls, objects, user=None, **kwargs):
        num_parent_objects, num_child_objects = 0, 0
        for page in objects:
            num_parent_objects += 1
            num_child_objects += page.get_descendant_count()
            page.delete(user=user)
        return num_parent_objects, num_child_objects

    def get_success_message(self, num_parent_objects, num_child_objects):
        if num_child_objects > 0:
            # Translators: This forms a message such as "1 page and 3 child pages have been deleted"
            return _("%(parent_pages)s and %(child_pages)s have been deleted") % {
                "parent_pages": self.get_parent_page_text(num_parent_objects),
                "child_pages": self.get_child_page_text(num_child_objects),
            }
        else:
            return ngettext(
                "%(num_parent_objects)d page has been deleted",
                "%(num_parent_objects)d pages have been deleted",
                num_parent_objects,
            ) % {"num_parent_objects": num_parent_objects}



# Generated by Django 1.9.1 on 2016-02-03 13:50
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0024_alter_page_content_type_on_delete_behaviour"),
        ("wagtailcore", "0027_fix_collection_path_collation"),
    ]

    operations = []



# Generated by Django 4.2.7 on 2024-02-12 21:04

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
        ("wagtailcore", "0092_alter_collectionviewrestriction_password_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="UploadedFile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("file", models.FileField(max_length=200, upload_to="wagtail_uploads")),
                (
                    "for_content_type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="uploads",
                        to="contenttypes.contenttype",
                        verbose_name="for content type",
                    ),
                ),
                (
                    "uploaded_by_user",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="uploaded by user",
                    ),
                ),
            ],
        ),
    ]



from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from wagtail import hooks
from wagtail.admin.menu import MenuItem
from wagtail.permissions import site_permission_policy

from .views import SiteViewSet


@hooks.register("register_admin_viewset")
def register_viewset():
    return SiteViewSet("wagtailsites", url_prefix="sites")


class SitesMenuItem(MenuItem):
    def is_shown(self, request):
        return site_permission_policy.user_has_any_permission(
            request.user, ["add", "change", "delete"]
        )


@hooks.register("register_settings_menu_item")
def register_sites_menu_item():
    return SitesMenuItem(
        _("Sites"),
        reverse("wagtailsites:index"),
        name="sites",
        icon_name="site",
        order=602,
    )



import re
from functools import lru_cache
from html import unescape

from django.core.validators import MaxLengthValidator
from django.db.models import Model
from django.template.loader import render_to_string
from django.utils.html import strip_tags
from django.utils.safestring import mark_safe

from wagtail.rich_text.feature_registry import FeatureRegistry
from wagtail.rich_text.rewriters import EmbedRewriter, LinkRewriter, MultiRuleRewriter

features = FeatureRegistry()


# Rewriter function to be built up on first call to expand_db_html, using the utility classes
# from wagtail.rich_text.rewriters along with the embed handlers / link handlers registered
# with the feature registry


@lru_cache(maxsize=None)
def get_rewriter():
    embed_rules = features.get_embed_types()
    link_rules = features.get_link_types()
    return MultiRuleRewriter(
        [
            LinkRewriter(
                bulk_rules={
                    linktype: handler.expand_db_attributes_many
                    for linktype, handler in link_rules.items()
                },
                reference_extractors={
                    linktype: handler.extract_references
                    for linktype, handler in link_rules.items()
                },
            ),
            EmbedRewriter(
                bulk_rules={
                    embedtype: handler.expand_db_attributes_many
                    for embedtype, handler in embed_rules.items()
                },
                reference_extractors={
                    embedtype: handler.extract_references
                    for embedtype, handler in embed_rules.items()
                },
            ),
        ]
    )


def expand_db_html(html):
    """
    Expand database-representation HTML into proper HTML usable on front-end templates
    """
    rewriter = get_rewriter()
    return rewriter(html)


def extract_references_from_rich_text(html):
    rewriter = get_rewriter()
    yield from rewriter.extract_references(html)


def get_text_for_indexing(richtext):
    """
    Return a plain text version of a rich text string, suitable for search indexing;
    like Django's strip_tags, but ensures that whitespace is left between block elements
    so that <p>hello</p><p>world</p> gives "hello world", not "helloworld".
    """
    # insert space after </p>, </h1> - </h6>, </li> and </blockquote> tags
    richtext = re.sub(
        r"(</(p|h\d|li|blockquote)>)", r"\1 ", richtext, flags=re.IGNORECASE
    )
    # also insert space after <br /> and <hr />
    richtext = re.sub(r"(<(br|hr)\s*/>)", r"\1 ", richtext, flags=re.IGNORECASE)
    return unescape(strip_tags(richtext).strip())


class RichText:
    """
    A custom object used to represent a renderable rich text value.
    Provides a 'source' property to access the original source code,
    and renders to the front-end HTML rendering.
    Used as the native value of a wagtailcore.blocks.field_block.RichTextBlock.
    """

    def __init__(self, source):
        self.source = source or ""

    def __html__(self):
        return render_to_string(
            "wagtailcore/shared/richtext.html", {"html": expand_db_html(self.source)}
        )

    def __str__(self):
        return mark_safe(self.__html__())

    def __bool__(self):
        return bool(self.source)

    def __eq__(self, other):
        if isinstance(other, RichText):
            return self.source == other.source
        return False


class EntityHandler:
    """
    An 'entity' is a placeholder tag within the saved rich text, which needs to be rewritten
    into real HTML at the point of rendering. Typically (but not necessarily) the entity will
    be a reference to a model to be fetched to have its data output into the rich text content
    (so that we aren't storing potentially changeable data within the saved rich text).

    An EntityHandler defines how this rewriting is performed.

    Currently Wagtail supports two kinds of entity: links (represented as <a linktype="...">...</a>)
    and embeds (represented as <embed embedtype="..." />).
    """

    @staticmethod
    def get_model():
        """
        If supported, returns the type of model able to be handled by this handler, e.g. Page.
        """
        raise NotImplementedError

    @classmethod
    def get_instance(cls, attrs: dict) -> Model:
        model = cls.get_model()
        return model._default_manager.get(id=attrs["id"])

    @classmethod
    def get_many(cls, attrs_list: list[dict]) -> list[Model]:
        model = cls.get_model()
        instance_ids = [attrs.get("id") for attrs in attrs_list]
        instances_by_id = model._default_manager.in_bulk(instance_ids)
        instances_by_str_id = {str(k): v for k, v in instances_by_id.items()}
        return [instances_by_str_id.get(str(id_)) for id_ in instance_ids]

    @staticmethod
    def expand_db_attributes(attrs: dict) -> str:
        """
        Given a dict of attributes from the entity tag
        stored in the database, returns the real HTML representation.
        """
        raise NotImplementedError

    @classmethod
    def expand_db_attributes_many(cls, attrs_list: list[dict]) -> list[str]:
        """
        Given a list of attribute dicts from a list of entity tags stored in
        the database, return the real HTML representation of each one.
        """
        return list(map(cls.expand_db_attributes, attrs_list))

    @classmethod
    def extract_references(cls, attrs):
        """
        Yields a sequence of (content_type_id, object_id, model_path, content_path) tuples for the
        database objects referenced by this entity, as per
        wagtail.models.ReferenceIndex._extract_references_from_object
        """
        return []


class LinkHandler(EntityHandler):
    pass


class EmbedHandler(EntityHandler):
    pass


class RichTextMaxLengthValidator(MaxLengthValidator):
    """
    A variant of MaxLengthValidator that only counts text (not HTML tags) towards the limit
    Un-escapes entities for consistency with client-side character count.
    """

    def clean(self, x):
        return len(unescape(strip_tags(x)))



import json
from datetime import date, datetime, timedelta
from io import BytesIO, StringIO

from django.contrib.auth.models import Permission
from django.core import management
from django.test import TestCase
from django.urls import reverse
from django.utils import timezone
from openpyxl import load_workbook

from wagtail.admin.admin_url_finder import AdminURLFinder
from wagtail.admin.tests.test_reports_views import BaseReportViewTestCase
from wagtail.contrib.search_promotions.models import (
    Query,
    QueryDailyHits,
    SearchPromotion,
)
from wagtail.contrib.search_promotions.templatetags.wagtailsearchpromotions_tags import (
    get_search_promotions,
)
from wagtail.log_actions import registry as log_registry
from wagtail.test.utils import WagtailTestUtils
from wagtail.test.utils.template_tests import AdminTemplateTestUtils


class TestSearchPromotions(TestCase):
    def test_search_pick_create(self):
        # Create a search pick to the root page
        SearchPromotion.objects.create(
            query=Query.get("root page"),
            page_id=1,
            sort_order=0,
            description="First search promotion",
        )

        # Check
        self.assertEqual(Query.get("root page").editors_picks.count(), 1)
        self.assertEqual(Query.get("root page").editors_picks.first().page_id, 1)

    def test_search_pick_link_create(self):
        # Add 3 search picks in a different order to their sort_order values
        # They should be ordered by their sort order values and not their insertion order
        SearchPromotion.objects.create(
            query=Query.get("root page"),
            external_link_url="https://wagtail.org",
            sort_order=0,
            description="First search promotion",
        )

        # Check
        self.assertEqual(Query.get("root page").editors_picks.count(), 1)
        self.assertEqual(
            Query.get("root page").editors_picks.first().external_link_url,
            "https://wagtail.org",
        )

    def test_search_pick_ordering(self):
        SearchPromotion.objects.create(
            query=Query.get("root page"),
            page_id=1,
            sort_order=0,
            description="First search pick",
        )
        SearchPromotion.objects.create(
            query=Query.get("root page"),
            page_id=1,
            sort_order=2,
            description="Last search pick",
        )
        SearchPromotion.objects.create(
            query=Query.get("root page"),
            external_link_url="https://wagtail.org",
            sort_order=1,
            description="Middle search pick",
        )

        # Check
        self.assertEqual(Query.get("root page").editors_picks.count(), 3)
        self.assertEqual(
            Query.get("root page").editors_picks.first().description,
            "First search pick",
        )
        self.assertEqual(
            Query.get("root page").editors_picks.last().description, "Last search pick"
        )

    def test_get_most_popular(self):
        popularQuery = Query.get("popular")
        for i in range(5):
            popularQuery.add_hit()
        SearchPromotion.objects.create(
            query=Query.get("popular"),
            page_id=2,
            sort_order=0,
            description="Popular search pick",
        )
        SearchPromotion.objects.create(
            query=Query.get("uninteresting"),
            page_id=1,
            sort_order=2,
            description="Uninteresting search pick",
        )

        # Check
        self.assertEqual(Query.get_most_popular().count(), 1)
        popular_picks = Query.get("popular").editors_picks.first()
        self.assertEqual(
            popular_picks.description,
            "Popular search pick",
        )
        self.assertEqual(popular_picks.query.hits, 5)

    def test_get_most_popular_since(self):
        TODAY = date.today()
        TWO_DAYS_AGO = TODAY - timedelta(days=2)
        FIVE_DAYS_AGO = TODAY - timedelta(days=5)

        popularQuery = Query.get("popular")
        for i in range(5):
            popularQuery.add_hit(date=FIVE_DAYS_AGO)

        surpriseQuery = Query.get("surprise")
        surpriseQuery.add_hit(date=TODAY)
        surpriseQuery.add_hit(date=TWO_DAYS_AGO)
        surpriseQuery.add_hit(date=FIVE_DAYS_AGO)
        SearchPromotion.objects.create(
            query=Query.get("popular"),
            page_id=2,
            sort_order=0,
            description="Popular search pick",
        )
        SearchPromotion.objects.create(
            query=Query.get("surprise"),
            page_id=2,
            sort_order=2,
            description="Surprising search pick",
        )

        # Check
        most_popular_queries = Query.get_most_popular(date_since=TWO_DAYS_AGO)
        self.assertEqual(most_popular_queries.count(), 1)
        editors_picks = Query.get("surprise").editors_picks
        surprise_picks = editors_picks.first()
        self.assertEqual(
            surprise_picks.description,
            "Surprising search pick",
        )
        self.assertEqual(surprise_picks.query.hits, 3)


class TestGetSearchPromotionsTemplateTag(TestCase):
    def test_get_search_promotions_template_tag(self):
        # Create a search pick to the root page
        pick = SearchPromotion.objects.create(
            query=Query.get("root page"),
            page_id=1,
            sort_order=0,
            description="First search pick",
        )

        # Create another search pick against a different query
        SearchPromotion.objects.create(
            query=Query.get("root page again"),
            page_id=1,
            sort_order=0,
            description="Second search pick",
        )

        # Check
        search_picks = list(get_search_promotions("root page"))
        self.assertEqual(search_picks, [pick])

    def test_get_search_promotions_with_none_query_string(self):
        search_picks = list(get_search_promotions(None))
        self.assertEqual(search_picks, [])


class TestSearchPromotionsIndexView(AdminTemplateTestUtils, WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()

    def test_simple(self):
        response = self.client.get(reverse("wagtailsearchpromotions:index"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/index.html")
        self.assertBreadcrumbsItemsRendered(
            [{"url": "", "label": "Promoted search results"}],
            response.content,
        )

    def test_search(self):
        response = self.client.get(
            reverse("wagtailsearchpromotions:index"), {"q": "Hello"}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["query_string"], "Hello")
        self.assertContains(
            response,
            'Sorry, no promoted results match "<em>Hello</em>"',
        )

    def test_search_with_results(self):
        SearchPromotion.objects.create(
            query=Query.get("search promotion query"),
            page_id=1,
        )
        SearchPromotion.objects.create(
            query=Query.get("search promotion query"),
            external_link_url="https://wagtail.org",
        )

        response = self.client.get(
            reverse("wagtailsearchpromotions:index"), {"q": "search promotion query"}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["query_string"], "search promotion query")
        self.assertContains(response, '<a href="/admin/pages/1/edit/" class="nolink">')
        self.assertContains(
            response,
            '<a href="https://wagtail.org" class="nolink" target="_blank" rel="noreferrer">',
        )

    def make_search_picks(self):
        for i in range(50):
            SearchPromotion.objects.create(
                query=Query.get("query " + str(i)),
                page_id=1,
                sort_order=0,
                description="First search pick",
            )

    def test_pagination(self):
        self.make_search_picks()

        response = self.client.get(reverse("wagtailsearchpromotions:index"), {"p": 2})

        # Check response
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/index.html")

        # Check that we got the correct page
        self.assertEqual(response.context["page_obj"].number, 2)

    def test_pagination_invalid(self):
        self.make_search_picks()

        response = self.client.get(
            reverse("wagtailsearchpromotions:index"), {"p": "Hello World!"}
        )

        # Check response
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/index.html")

        # Check that we got page one
        self.assertEqual(response.context["page_obj"].number, 1)

    def test_pagination_out_of_range(self):
        self.make_search_picks()

        response = self.client.get(
            reverse("wagtailsearchpromotions:index"), {"p": 99999}
        )

        # Check response
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/index.html")

        # Check that we got the last page
        self.assertEqual(
            response.context["page_obj"].number,
            response.context["paginator"].num_pages,
        )

    def test_num_queries(self):
        url = reverse("wagtailsearchpromotions:index")
        self.make_search_picks()
        # Warm up the cache
        self.client.get(url)

        # Number of queries with the current number of search picks
        with self.assertNumQueries(11):
            self.client.get(url)

        # Add more SearchPromotions and QueryDailyHits to some of the queries
        today = date.today()
        for i in range(20):
            query = Query.get("query " + str(i))
            promos = [
                SearchPromotion(
                    query=query,
                    page_id=j % 2 + 1,
                    sort_order=j,
                    description=f"Search pick {j}",
                )
                for j in range(5)
            ]
            hits = [
                QueryDailyHits(query=query, date=today - timedelta(days=j), hits=j)
                for j in range(5)
            ]
            SearchPromotion.objects.bulk_create(promos)
            QueryDailyHits.objects.bulk_create(hits)

        # Number of queries after the addition of more search picks and hits
        # should remain the same (no N+1 queries)
        with self.assertNumQueries(11):
            self.client.get(url)

    def test_results_are_ordered_alphabetically(self):
        self.make_search_picks()
        SearchPromotion.objects.create(
            query=Query.get("aaargh snake"),
            page_id=1,
            sort_order=0,
            description="ooh, it's a snake",
        )
        # Add another one to make sure it's not ordered descending by pk
        SearchPromotion.objects.create(
            query=Query.get("beloved snake"),
            page_id=1,
            sort_order=0,
            description="beloved snake goes ssSSSS",
        )

        response = self.client.get(reverse("wagtailsearchpromotions:index"))

        # Check response
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/index.html")

        # "aargh snake" should be the first result alphabetically
        self.assertEqual(response.context["queries"][0].query_string, "aaargh snake")
        self.assertEqual(response.context["queries"][1].query_string, "beloved snake")

    def test_multiple_searchpromotions(self):
        today = date.today()
        for i in range(10):
            Query.get("root page").add_hit(date=today - timedelta(days=i))
        SearchPromotion.objects.create(
            query=Query.get("root page"),
            page_id=1,
            sort_order=0,
            description="First search pick",
        )
        SearchPromotion.objects.create(
            query=Query.get("root page"),
            page_id=2,
            sort_order=0,
            description="Second search pick",
        )
        response = self.client.get(reverse("wagtailsearchpromotions:index"))

        self.assertContains(response, "<td>10</td>", html=True)
        self.assertEqual(Query.get("root page").hits, 10)

        soup = self.get_soup(response.content)
        root_page_edit_url = reverse("wagtailadmin_pages:edit", args=(1,))
        homepage_edit_url = reverse("wagtailadmin_pages:edit", args=(2,))
        root_page_edit_link = soup.select_one(f'a[href="{root_page_edit_url}"]')
        homepage_edit_link = soup.select_one(f'a[href="{homepage_edit_url}"]')
        self.assertIsNotNone(root_page_edit_link)
        self.assertIsNotNone(homepage_edit_link)
        self.assertEqual(Query.get("root page").editors_picks.count(), 2)

    def test_results_ordering(self):
        self.make_search_picks()
        url = reverse("wagtailsearchpromotions:index")
        SearchPromotion.objects.create(
            query=Query.get("zyzzyvas"),
            page_id=1,
            sort_order=0,
            description="no definition found, this is a valid scrabble word though",
        )

        SearchPromotion.objects.create(
            query=Query.get("aardwolf"),
            page_id=1,
            sort_order=22,
            description="Striped hyena of southeast Africa that feeds chiefly on insects",
        )

        popularQuery = Query.get("optimal")
        for i in range(50):
            popularQuery.add_hit()
        SearchPromotion.objects.create(
            query=popularQuery,
            page_id=1,
            sort_order=15,
            description="An oddly popular search term?",
        )

        popularQuery = Query.get("suboptimal")
        for i in range(25):
            popularQuery.add_hit()
        SearchPromotion.objects.create(
            query=popularQuery,
            page_id=1,
            sort_order=15,
            description="Not as popular",
        )

        # ordered by querystring (reversed)
        response = self.client.get(url + "?ordering=-query_string")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["page_obj"][0].query_string, "zyzzyvas")

        # last page, still ordered by query string (reversed)
        response = self.client.get(url + "?ordering=-query_string&p=3")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["page_obj"][-1].query_string, "aardwolf")

        # ordered by querystring (not reversed)
        response = self.client.get(url + "?ordering=query_string")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["page_obj"][0].query_string, "aardwolf")

        # ordered by sum of daily hits (reversed)
        response = self.client.get(url + "?ordering=-views")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["page_obj"][0].query_string, "optimal")
        self.assertEqual(response.context["page_obj"][1].query_string, "suboptimal")

        # ordered by sum of daily hits, last page (not reversed)
        response = self.client.get(url + "?ordering=views&p=3")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["page_obj"][-1].query_string, "optimal")
        self.assertEqual(response.context["page_obj"][-2].query_string, "suboptimal")

    def test_get_with_no_permission(self):
        self.user.is_superuser = False
        self.user.save()
        # Only basic access_admin permission is given
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin",
                codename="access_admin",
            )
        )

        response = self.client.get(reverse("wagtailsearchpromotions:index"))
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_get_with_edit_permission_only(self):
        self.user.is_superuser = False
        self.user.save()
        # Only basic access_admin permission is given
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin",
                codename="access_admin",
            ),
            Permission.objects.get(
                content_type__app_label="wagtailsearchpromotions",
                codename="change_searchpromotion",
            ),
        )

        response = self.client.get(reverse("wagtailsearchpromotions:index"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/index.html")

        soup = self.get_soup(response.content)
        add_url = reverse("wagtailsearchpromotions:add")
        # Should not render add link
        self.assertIsNone(soup.select_one(f'a[href="{add_url}"]'))


class TestSearchPromotionsAddView(AdminTemplateTestUtils, WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()

    def test_simple(self):
        response = self.client.get(reverse("wagtailsearchpromotions:add"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/add.html")
        self.assertBreadcrumbsItemsRendered(
            [
                {
                    "url": reverse("wagtailsearchpromotions:index"),
                    "label": "Promoted search results",
                },
                {"url": "", "label": "New: Promoted search result"},
            ],
            response.content,
        )

    def test_post(self):
        # Submit
        post_data = {
            "query_string": "test",
            "editors_picks-TOTAL_FORMS": 1,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Hello",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be redirected back to the index
        self.assertRedirects(response, reverse("wagtailsearchpromotions:index"))

        # Check that the search pick was created
        self.assertTrue(Query.get("test").editors_picks.filter(page_id=1).exists())

        # Ensure that only one log entry was created for the search pick
        search_picks = list(Query.get("test").editors_picks.all())
        self.assertEqual(len(search_picks), 1)
        self.assertTrue(search_picks[0].page_id, 1)
        logs = log_registry.get_logs_for_instance(search_picks[0])
        self.assertEqual(len(logs), 1)
        self.assertEqual(logs[0].action, "wagtail.create")

    def test_with_multiple_picks(self):
        # Submit
        post_data = {
            "query_string": "test",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Hello",
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-page": "",
            "editors_picks-1-external_link_url": "https://wagtail.org",
            "editors_picks-1-external_link_text": "Wagtail",
            "editors_picks-1-description": "The landing page",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be redirected back to the index
        self.assertRedirects(response, reverse("wagtailsearchpromotions:index"))

        # Check that the search pick was created
        search_picks = list(
            Query.get("test").editors_picks.all().order_by("description")
        )
        self.assertEqual(len(search_picks), 2)
        self.assertEqual(search_picks[0].page_id, 1)
        self.assertEqual(search_picks[0].description, "Hello")
        self.assertEqual(search_picks[1].external_link_url, "https://wagtail.org")
        self.assertEqual(search_picks[1].description, "The landing page")

        # Ensure that only one log entry was created for each search pick
        for search_pick in search_picks:
            logs = log_registry.get_logs_for_instance(search_pick)
            self.assertEqual(len(logs), 1)
            self.assertEqual(logs[0].action, "wagtail.create")
            self.assertEqual(logs[0].user, self.user)

    def test_post_with_existing_query_string(self):
        # Create an existing query with search picks
        query = Query.get("test")
        search_pick_1 = query.editors_picks.create(
            page_id=1, sort_order=0, description="Root page"
        )
        search_pick_2 = query.editors_picks.create(
            page_id=2, sort_order=1, description="Homepage"
        )

        # Submit
        post_data = {
            "query_string": "test",
            "editors_picks-TOTAL_FORMS": 1,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 1,
            "editors_picks-0-external_link_url": "https://wagtail.org",
            "editors_picks-0-external_link_text": "Wagtail",
            "editors_picks-0-description": "A Django-based CMS",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be redirected back to the index
        self.assertRedirects(response, reverse("wagtailsearchpromotions:index"))

        # Check that the submitted search pick is created
        # and the existing ones are still there
        self.assertEqual(
            set(
                Query.get("test")
                .editors_picks.all()
                .values_list("page_id", "external_link_url")
            ),
            {
                (search_pick_1.page_id, ""),
                (search_pick_2.page_id, ""),
                (None, "https://wagtail.org"),
            },
        )

    def test_post_with_invalid_query_string(self):
        # Submit
        post_data = {
            "query_string": "",
            "editors_picks-TOTAL_FORMS": 1,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Hello",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be given an error on the specific field in the form
        self.assertEqual(response.status_code, 200)

        self.assertFormError(
            response.context["form"], "query_string", "This field is required."
        )
        # The formset should still contain the submitted data
        self.assertEqual(len(response.context["searchpicks_formset"].forms), 1)
        self.assertEqual(
            response.context["searchpicks_formset"].forms[0].cleaned_data["page"].id,
            1,
        )
        self.assertEqual(
            response.context["searchpicks_formset"]
            .forms[0]
            .cleaned_data["description"],
            "Hello",
        )
        # Should not raise an error anywhere else
        self.assertFormSetError(response.context["searchpicks_formset"], 0, "page", [])
        self.assertFormSetError(response.context["searchpicks_formset"], 0, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])

    def test_post_with_invalid_page(self):
        # Submit
        post_data = {
            "query_string": "test",
            "editors_picks-TOTAL_FORMS": 1,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 9999999999,
            "editors_picks-0-description": "Hello",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be given an error on the specific field in the form
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            0,
            "page",
            "Select a valid choice. "
            "That choice is not one of the available choices.",
        )
        # Should not raise an error anywhere else
        self.assertFormSetError(response.context["searchpicks_formset"], 0, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])

    def test_post_with_external_link(self):
        # Submit
        post_data = {
            "query_string": "test",
            "editors_picks-TOTAL_FORMS": 1,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-external_link_url": "https://wagtail.org",
            "editors_picks-0-external_link_text": "Wagtail",
            "editors_picks-0-description": "Hello",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be redirected back to the index
        self.assertRedirects(response, reverse("wagtailsearchpromotions:index"))

        # Check that the search pick was created
        self.assertTrue(
            Query.get("test")
            .editors_picks.filter(external_link_url="https://wagtail.org")
            .exists()
        )

    def test_post_without_recommendations(self):
        # Submit
        post_data = {
            "query_string": "test",
            "editors_picks-TOTAL_FORMS": 0,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be given an error
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            None,
            None,
            "Please specify at least one recommendation for this search term.",
        )

    def test_post_with_page_and_external_link(self):
        post_data = {
            "query_string": "test",
            "editors_picks-TOTAL_FORMS": 1,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-external_link_url": "https://wagtail.org",
            "editors_picks-0-external_link_text": "Wagtail",
            "editors_picks-0-description": "Hello",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be given an error on a specific form in the formset
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            0,
            None,
            "Please only select a page OR enter an external link.",
        )
        # Should not raise an error on the top-level formset
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])

    def test_post_missing_recommendation(self):
        post_data = {
            "query_string": "test",
            "editors_picks-TOTAL_FORMS": 1,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-description": "Hello",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be given an error on a specific form in the formset
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            0,
            None,
            "You must recommend a page OR an external link.",
        )
        # Should not raise an error on the top-level formset
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])

    def test_post_invalid_external_link(self):
        post_data = {
            "query_string": "test",
            "editors_picks-TOTAL_FORMS": 1,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-external_link_url": "notalink",
            "editors_picks-0-external_link_text": "Wagtail",
            "editors_picks-0-description": "Hello",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be given an error on the specific field in the form
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            0,
            "external_link_url",
            "Enter a valid URL.",
        )
        # Should not raise an error anywhere else
        self.assertFormSetError(response.context["searchpicks_formset"], 0, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])

    def test_post_missing_external_text(self):
        post_data = {
            "query_string": "test",
            "editors_picks-TOTAL_FORMS": 1,
            "editors_picks-INITIAL_FORMS": 0,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-external_link_url": "https://wagtail.org",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be given an error on the specific field in the form
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            0,
            "external_link_text",
            "You must enter an external link text if you enter an external link URL.",
        )

        # Should not raise an error anywhere else
        self.assertFormSetError(response.context["searchpicks_formset"], 0, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])

    def test_get_with_no_permission(self):
        self.user.is_superuser = False
        self.user.save()
        # Only basic access_admin permission is given
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin",
                codename="access_admin",
            )
        )

        response = self.client.get(reverse("wagtailsearchpromotions:add"))
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_get_with_add_permission_only(self):
        self.user.is_superuser = False
        self.user.save()
        # Only basic access_admin permission is given
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin",
                codename="access_admin",
            ),
            Permission.objects.get(
                content_type__app_label="wagtailsearchpromotions",
                codename="add_searchpromotion",
            ),
        )

        response = self.client.get(reverse("wagtailsearchpromotions:add"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/add.html")


class TestSearchPromotionsEditView(AdminTemplateTestUtils, WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()

        # Create a search pick to edit
        self.query = Query.get("Hello")
        self.search_pick = self.query.editors_picks.create(
            page_id=1, sort_order=0, description="Root page"
        )
        self.search_pick_2 = self.query.editors_picks.create(
            page_id=2, sort_order=1, description="Homepage"
        )

    def test_simple(self):
        response = self.client.get(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,))
        )
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/edit.html")

        url_finder = AdminURLFinder(self.user)
        expected_url = "/admin/searchpicks/%d/" % self.query.id
        self.assertEqual(url_finder.get_edit_url(self.search_pick), expected_url)

        self.assertBreadcrumbsItemsRendered(
            [
                {
                    "url": reverse("wagtailsearchpromotions:index"),
                    "label": "Promoted search results",
                },
                {"url": "", "label": "hello"},
            ],
            response.content,
        )

    def test_post(self):
        # Submit
        post_data = {
            "query_string": "Hello",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Description has changed",  # Change
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-page": 2,
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be redirected back to the index
        self.assertRedirects(response, reverse("wagtailsearchpromotions:index"))

        # Check that the search pick description was edited
        self.assertEqual(
            SearchPromotion.objects.get(id=self.search_pick.id).description,
            "Description has changed",
        )

        search_picks = list(
            Query.get("Hello").editors_picks.all().order_by("description")
        )
        self.assertEqual(len(search_picks), 2)
        self.assertEqual(search_picks[0].page_id, 1)
        self.assertEqual(search_picks[0].description, "Description has changed")
        self.assertEqual(search_picks[1].page_id, 2)
        self.assertEqual(search_picks[1].description, "Homepage")

        # Ensure that only one log entry was created for each search pick
        for search_pick in search_picks:
            logs = log_registry.get_logs_for_instance(search_pick)
            self.assertEqual(len(logs), 1)
            self.assertEqual(logs[0].action, "wagtail.edit")
            self.assertEqual(logs[0].user, self.user)

    def test_post_with_invalid_query_string(self):
        # Submit
        post_data = {
            "query_string": "",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Description has changed",  # Change
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-page": 2,
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(reverse("wagtailsearchpromotions:add"), post_data)

        # User should be given an error on the specific field in the form
        self.assertEqual(response.status_code, 200)
        self.assertFormError(
            response.context["form"], "query_string", "This field is required."
        )
        # The formset should still contain the submitted data
        self.assertEqual(len(response.context["searchpicks_formset"].forms), 2)
        self.assertEqual(
            response.context["searchpicks_formset"].forms[0].cleaned_data["page"].id,
            1,
        )
        self.assertEqual(
            response.context["searchpicks_formset"]
            .forms[0]
            .cleaned_data["description"],
            "Description has changed",
        )
        # Should not raise an error anywhere else
        self.assertFormSetError(response.context["searchpicks_formset"], 0, "page", [])
        self.assertFormSetError(response.context["searchpicks_formset"], 0, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])

    def test_post_with_invalid_page(self):
        # Submit
        post_data = {
            "query_string": "Hello",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Description has changed",  # Change
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-page": 9214599,
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be given an error on the specific field in the form
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            1,
            "page",
            "Select a valid choice. "
            "That choice is not one of the available choices.",
        )
        # Should not raise an error anywhere else
        self.assertFormSetError(response.context["searchpicks_formset"], 0, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], 1, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])

    def test_post_change_query_string(self):
        current_picks = set(self.query.editors_picks.all())
        # Submit
        post_data = {
            "query_string": "Hello again",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Description has changed",  # Change
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-page": 2,
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be redirected back to the index
        self.assertRedirects(response, reverse("wagtailsearchpromotions:index"))

        # Ensure search picks from the old query are moved to the new one
        new_query = Query.get("Hello again")
        self.assertEqual(set(new_query.editors_picks.all()), current_picks)
        self.search_pick.refresh_from_db()
        self.assertEqual(self.search_pick.query, new_query)
        self.assertEqual(self.query.editors_picks.count(), 0)

        # Check that the search pick description was edited
        self.assertEqual(self.search_pick.description, "Description has changed")

    def test_post_reorder(self):
        # Check order before reordering
        self.assertEqual(Query.get("Hello").editors_picks.all()[0], self.search_pick)
        self.assertEqual(Query.get("Hello").editors_picks.all()[1], self.search_pick_2)

        # Submit
        post_data = {
            "query_string": "Hello",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 1,  # Change
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Root page",
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 0,  # Change
            "editors_picks-1-page": 2,
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be redirected back to the index
        self.assertRedirects(response, reverse("wagtailsearchpromotions:index"))

        # Check that the ordering has been saved correctly
        self.assertEqual(
            SearchPromotion.objects.get(id=self.search_pick.id).sort_order, 1
        )
        self.assertEqual(
            SearchPromotion.objects.get(id=self.search_pick_2.id).sort_order, 0
        )

        # Check that the recommendations were reordered
        self.assertEqual(Query.get("Hello").editors_picks.all()[0], self.search_pick_2)
        self.assertEqual(Query.get("Hello").editors_picks.all()[1], self.search_pick)

    def test_post_with_external_link(self):
        # Submit
        post_data = {
            "query_string": "Hello",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 1,  # Change
            "editors_picks-0-external_link_url": "https://wagtail.org",
            "editors_picks-0-external_link_text": "Wagtail",
            "editors_picks-0-description": "Root page",
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 0,  # Change
            "editors_picks-1-external_link_url": "https://djangoproject.com",
            "editors_picks-1-external_link_text": "Django",
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be redirected back to the index
        self.assertRedirects(response, reverse("wagtailsearchpromotions:index"))

        # Check that the search pick was created
        self.assertTrue(
            Query.get("Hello")
            .editors_picks.filter(external_link_url="https://wagtail.org")
            .exists()
        )
        self.assertTrue(
            Query.get("Hello")
            .editors_picks.filter(external_link_url="https://djangoproject.com")
            .exists()
        )

    def test_post_delete_recommendation(self):
        # Submit
        post_data = {
            "query_string": "Hello",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Root page",
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": 1,
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-page": 2,
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be redirected back to the index
        self.assertRedirects(response, reverse("wagtailsearchpromotions:index"))

        # Check that the recommendation was deleted
        self.assertFalse(
            SearchPromotion.objects.filter(id=self.search_pick_2.id).exists()
        )

        # The other recommendation should still exist
        self.assertTrue(SearchPromotion.objects.filter(id=self.search_pick.id).exists())

    def test_post_without_recommendations(self):
        # Submit
        post_data = {
            "query_string": "Hello",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": 1,
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Description has changed",  # Change
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": 1,
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-page": 2,
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be given an error
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            None,
            None,
            "Please specify at least one recommendation for this search term.",
        )

    def test_post_with_page_and_external_link(self):
        post_data = {
            "query_string": "Hello",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Description has changed",  # Change
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-page": 2,
            "editors_picks-1-external_link_url": "https://wagtail.org",
            "editors_picks-1-external_link_text": "Wagtail",
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be given an error on a specific form in the formset
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            1,
            None,
            "Please only select a page OR enter an external link.",
        )
        # Should not raise an error anywhere else
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], 0, None, [])

    def test_post_missing_recommendation(self):
        post_data = {
            "query_string": "Hello",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-description": "Description has changed",  # Change
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-external_link_url": "https://wagtail.org",
            "editors_picks-1-external_link_text": "Wagtail",
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be given an error on a specific form in the formset
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            0,
            None,
            "You must recommend a page OR an external link.",
        )
        # Should not raise an error anywhere else
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], 1, None, [])

    def test_post_invalid_external_link(self):
        post_data = {
            "query_string": "Hello",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Description has changed",  # Change
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-external_link_url": "notalink",
            "editors_picks-1-external_link_text": "Wagtail",
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be given an error on the specific field in the form
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            1,
            "external_link_url",
            "Enter a valid URL.",
        )
        # Should not raise an error anywhere else
        self.assertFormSetError(response.context["searchpicks_formset"], 1, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])

    def test_post_missing_external_text(self):
        post_data = {
            "query_string": "Hello",
            "editors_picks-TOTAL_FORMS": 2,
            "editors_picks-INITIAL_FORMS": 2,
            "editors_picks-MAX_NUM_FORMS": 1000,
            "editors_picks-0-id": self.search_pick.id,
            "editors_picks-0-DELETE": "",
            "editors_picks-0-ORDER": 0,
            "editors_picks-0-page": 1,
            "editors_picks-0-description": "Description has changed",  # Change
            "editors_picks-1-id": self.search_pick_2.id,
            "editors_picks-1-DELETE": "",
            "editors_picks-1-ORDER": 1,
            "editors_picks-1-external_link_url": "https://wagtail.org",
            "editors_picks-1-description": "Homepage",
        }
        response = self.client.post(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)), post_data
        )

        # User should be given an error on the specific field in the form
        self.assertEqual(response.status_code, 200)
        self.assertFormSetError(
            response.context["searchpicks_formset"],
            1,
            "external_link_text",
            "You must enter an external link text if you enter an external link URL.",
        )

        # Should not raise an error anywhere else
        self.assertFormSetError(response.context["searchpicks_formset"], 1, None, [])
        self.assertFormSetError(response.context["searchpicks_formset"], None, None, [])

    def test_get_with_no_permission(self):
        self.user.is_superuser = False
        self.user.save()
        # Only basic access_admin permission is given
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin",
                codename="access_admin",
            )
        )

        response = self.client.get(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)),
        )
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_get_with_edit_permission_only(self):
        self.user.is_superuser = False
        self.user.save()
        # Only basic access_admin permission is given
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin",
                codename="access_admin",
            ),
            Permission.objects.get(
                content_type__app_label="wagtailsearchpromotions",
                codename="change_searchpromotion",
            ),
        )

        response = self.client.get(
            reverse("wagtailsearchpromotions:edit", args=(self.query.id,)),
        )
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/edit.html")


class TestSearchPromotionsDeleteView(WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()

        # Create a search pick to delete
        self.query = Query.get("Hello")
        self.search_pick = self.query.editors_picks.create(
            page_id=1, description="Root page"
        )
        self.search_pick_2 = self.query.editors_picks.create(
            page_id=2, description="Homepage"
        )

    def test_simple(self):
        response = self.client.get(
            reverse("wagtailsearchpromotions:delete", args=(self.query.id,))
        )
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/confirm_delete.html")

    def test_post(self):
        # Submit
        response = self.client.post(
            reverse("wagtailsearchpromotions:delete", args=(self.query.id,))
        )

        # User should be redirected back to the index
        self.assertRedirects(response, reverse("wagtailsearchpromotions:index"))

        # Check that both recommendations were deleted
        self.assertFalse(
            SearchPromotion.objects.filter(id=self.search_pick_2.id).exists()
        )

        # The other recommendation should still exist
        self.assertFalse(
            SearchPromotion.objects.filter(id=self.search_pick.id).exists()
        )

    def test_get_with_no_permission(self):
        self.user.is_superuser = False
        self.user.save()
        # Only basic access_admin permission is given
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin",
                codename="access_admin",
            )
        )

        response = self.client.get(
            reverse("wagtailsearchpromotions:delete", args=(self.query.id,)),
        )
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_get_with_edit_permission_only(self):
        self.user.is_superuser = False
        self.user.save()
        # Only basic access_admin permission is given
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin",
                codename="access_admin",
            ),
            Permission.objects.get(
                content_type__app_label="wagtailsearchpromotions",
                codename="delete_searchpromotion",
            ),
        )

        response = self.client.get(
            reverse("wagtailsearchpromotions:delete", args=(self.query.id,)),
        )
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailsearchpromotions/confirm_delete.html")


class TestGarbageCollectManagementCommand(TestCase):
    def test_garbage_collect_command(self):
        nowdt = datetime.now()
        old_hit_date = (nowdt - timedelta(days=14)).date()
        recent_hit_date = (nowdt - timedelta(days=1)).date()

        # Add 10 hits that are more than one week old. The related queries and the daily hits
        # should be deleted by the search_garbage_collect command.
        query_ids_to_be_deleted = []
        for i in range(10):
            q = Query.get(f"Hello {i}")
            q.add_hit(date=old_hit_date)
            query_ids_to_be_deleted.append(q.id)

        # Add 10 hits that are less than one week old. These ones should not be deleted.
        recent_query_ids = []
        for i in range(10):
            q = Query.get(f"World {i}")
            q.add_hit(date=recent_hit_date)
            recent_query_ids.append(q.id)

        # Add 10 queries that are promoted. These ones should not be deleted.
        promoted_query_ids = []
        for i in range(10):
            q = Query.get(f"Foo bar {i}")
            q.add_hit(date=old_hit_date)
            SearchPromotion.objects.create(
                query=q, page_id=1, sort_order=0, description="Test"
            )
            promoted_query_ids.append(q.id)

        management.call_command("searchpromotions_garbage_collect", stdout=StringIO())

        self.assertFalse(Query.objects.filter(id__in=query_ids_to_be_deleted).exists())
        self.assertFalse(
            QueryDailyHits.objects.filter(
                date=old_hit_date, query_id__in=query_ids_to_be_deleted
            ).exists()
        )

        self.assertEqual(Query.objects.filter(id__in=recent_query_ids).count(), 10)
        self.assertEqual(
            QueryDailyHits.objects.filter(
                date=recent_hit_date, query_id__in=recent_query_ids
            ).count(),
            10,
        )

        self.assertEqual(Query.objects.filter(id__in=promoted_query_ids).count(), 10)
        self.assertEqual(
            QueryDailyHits.objects.filter(
                date=recent_hit_date, query_id__in=promoted_query_ids
            ).count(),
            0,
        )


class TestQueryChooserView(WagtailTestUtils, TestCase):
    def setUp(self):
        self.login()

    def get(self, params={}):
        return self.client.get("/admin/searchpicks/queries/chooser/", params)

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtailsearchpromotions/queries/chooser/chooser.html"
        )
        response_json = json.loads(response.content.decode())
        self.assertEqual(response_json["step"], "chooser")

    def test_search(self):
        response = self.get({"q": "Hello"})
        self.assertEqual(response.status_code, 200)

    def test_pagination(self):
        # page numbers in range should be accepted
        response = self.get({"p": 1})
        self.assertEqual(response.status_code, 200)
        # page numbers out of range should return 404
        response = self.get({"p": 9999})
        self.assertEqual(response.status_code, 404)


class TestHitCounter(TestCase):
    def test_no_hits(self):
        self.assertEqual(Query.get("Hello").hits, 0)

    def test_hit(self):
        # Add a hit
        Query.get("Hello").add_hit()

        # Test
        self.assertEqual(Query.get("Hello").hits, 1)

    def test_10_hits(self):
        # Add 10 hits
        for i in range(10):
            Query.get("Hello").add_hit()

        # Test
        self.assertEqual(Query.get("Hello").hits, 10)


class TestQueryStringNormalisation(TestCase):
    def setUp(self):
        self.query = Query.get("  Hello  World!  ")

    def test_normalisation(self):
        self.assertEqual(str(self.query), "hello world!")

    def test_equivalent_queries(self):
        queries = [
            "  Hello World!",
            "Hello World!  ",
            "hello  world!",
            "  Hello  world!  ",
        ]

        for query in queries:
            self.assertEqual(self.query, Query.get(query))

    def test_different_queries(self):
        queries = [
            "HelloWorld",
            "HelloWorld!" "  Hello  World!  ",
            "Hello",
        ]

        for query in queries:
            self.assertNotEqual(self.query, Query.get(query))


class TestQueryPopularity(TestCase):
    def test_query_popularity(self):
        # Add 3 hits to unpopular query
        for i in range(3):
            Query.get("unpopular query").add_hit()

        # Add 10 hits to popular query
        for i in range(10):
            Query.get("popular query").add_hit()

        # Get most popular queries
        popular_queries = Query.get_most_popular()

        # Check list
        self.assertEqual(popular_queries.count(), 2)
        self.assertEqual(popular_queries[0], Query.get("popular query"))
        self.assertEqual(popular_queries[1], Query.get("unpopular query"))

        # Add 5 hits to little popular query
        for i in range(5):
            Query.get("little popular query").add_hit()

        # Check list again, little popular query should be in the middle
        self.assertEqual(popular_queries.count(), 3)
        self.assertEqual(popular_queries[0], Query.get("popular query"))
        self.assertEqual(popular_queries[1], Query.get("little popular query"))
        self.assertEqual(popular_queries[2], Query.get("unpopular query"))

        # Unpopular query goes viral!
        for i in range(20):
            Query.get("unpopular query").add_hit()

        # Unpopular query should be most popular now
        self.assertEqual(popular_queries.count(), 3)
        self.assertEqual(popular_queries[0], Query.get("unpopular query"))
        self.assertEqual(popular_queries[1], Query.get("popular query"))
        self.assertEqual(popular_queries[2], Query.get("little popular query"))


class TestQueryHitsReportView(BaseReportViewTestCase):
    url_name = "wagtailsearchpromotions:search_terms"

    @classmethod
    def setUpTestData(self):
        self.query = Query.get("A query with three hits")
        self.query.add_hit()
        self.query.add_hit()
        self.query.add_hit()
        Query.get("a query with no hits")
        Query.get("A query with one hit").add_hit()
        query = Query.get("A query with two hits")
        query.add_hit()
        query.add_hit()

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailadmin/reports/base_report.html")
        self.assertTemplateUsed(
            response,
            "wagtailadmin/reports/base_report_results.html",
        )
        self.assertBreadcrumbs(
            [{"url": "", "label": "Search terms"}],
            response.content,
        )

        soup = self.get_soup(response.content)
        trs = soup.select("main tr")

        # Default ordering should be by hits descending
        self.assertEqual(
            [[cell.text.strip() for cell in tr.select("th,td")] for tr in trs],
            [
                ["Search term(s)", "Views"],
                ["a query with three hits", "3"],
                ["a query with two hits", "2"],
                ["a query with one hit", "1"],
            ],
        )

        self.assertNotContains(response, "There are no results.")
        self.assertActiveFilterNotRendered(soup)
        self.assertPageTitle(soup, "Search terms - Wagtail")

    def test_get_with_no_permissions(self):
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )

        response = self.get()

        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_csv_export(self):
        response = self.get(params={"export": "csv"})
        self.assertEqual(response.status_code, 200)

        data_lines = response.getvalue().decode().splitlines()
        self.assertEqual(
            data_lines,
            [
                "Search term(s),Views",
                "a query with three hits,3",
                "a query with two hits,2",
                "a query with one hit,1",
            ],
        )

    def test_xlsx_export(self):
        response = self.get(params={"export": "xlsx"})
        self.assertEqual(response.status_code, 200)
        workbook_data = response.getvalue()
        worksheet = load_workbook(filename=BytesIO(workbook_data))["Sheet1"]
        cell_array = [[cell.value for cell in row] for row in worksheet.rows]
        self.assertEqual(
            cell_array,
            [
                ["Search term(s)", "Views"],
                ["a query with three hits", 3],
                ["a query with two hits", 2],
                ["a query with one hit", 1],
            ],
        )

    def test_ordering(self):
        cases = {
            "query_string": [
                ["a query with one hit", "1"],
                ["a query with three hits", "3"],
                ["a query with two hits", "2"],
            ],
            "-query_string": [
                ["a query with two hits", "2"],
                ["a query with three hits", "3"],
                ["a query with one hit", "1"],
            ],
            "_hits": [
                ["a query with one hit", "1"],
                ["a query with two hits", "2"],
                ["a query with three hits", "3"],
            ],
            "-_hits": [
                ["a query with three hits", "3"],
                ["a query with two hits", "2"],
                ["a query with one hit", "1"],
            ],
        }
        for ordering, results in cases.items():
            with self.subTest(ordering=ordering):
                response = self.get(params={"ordering": ordering})
                self.assertEqual(response.status_code, 200)
                soup = self.get_soup(response.content)
                trs = soup.select("main tbody tr")
                self.assertEqual(
                    [[cell.text.strip() for cell in tr.select("td")] for tr in trs],
                    results,
                )


class TestFilteredQueryHitsView(BaseReportViewTestCase):
    url_name = "wagtailsearchpromotions:search_terms"

    def setUp(self):
        self.user = self.login()
        self.query_hit = Query.get("This will be found")
        self.date = timezone.now().date()
        self.query_hit.add_hit(date=self.date)

    def test_search_by_query_string(self):
        response = self.get(params={"q": "Found"})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "this will be found")
        self.assertNotContains(response, "There are no results.")
        self.assertActiveFilterNotRendered(self.get_soup(response.content))

        response = self.get(params={"q": "Not found"})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "There are no results.")
        self.assertNotContains(response, "this will be found")
        self.assertActiveFilterNotRendered(self.get_soup(response.content))

    def test_filter_by_date(self):
        params = {
            "hit_date_from": self.date.replace(day=1, month=1),
        }
        response = self.get(params=params)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "this will be found")
        self.assertNotContains(response, "There are no results.")
        self.assertActiveFilter(
            self.get_soup(response.content), "hit_date_from", params["hit_date_from"]
        )

        params["hit_date_from"] = self.date.replace(year=self.date.year + 1)
        params["hit_date_to"] = self.date.replace(year=self.date.year + 2)

        response = self.get(params=params)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "There are no results.")
        self.assertNotContains(response, "this will be found")



from warnings import warn

from django.utils.translation import gettext_lazy as _

from wagtail.admin.views.generic import BaseListingView, PermissionCheckedMixin
from wagtail.admin.views.mixins import SpreadsheetExportMixin
from wagtail.permissions import page_permission_policy
from wagtail.utils.deprecation import RemovedInWagtail70Warning


class ReportView(SpreadsheetExportMixin, PermissionCheckedMixin, BaseListingView):
    template_name = "wagtailadmin/reports/base_report.html"
    results_template_name = "wagtailadmin/reports/base_report_results.html"
    title = ""
    paginate_by = 50

    def get_page_title(self):
        if self.page_title:
            return self.page_title
        # WagtailAdminTemplateMixin uses `page_title`, but the documented approach
        # for ReportView used `title`, so we need to support both during the
        # deprecation period. When `title` is removed, this and the `get_context_data`
        # overrides can be removed.
        warn(
            f"The `title` attribute in `{self.__class__.__name__}` (a `ReportView` subclass) "
            "is deprecated. Use `page_title` instead.",
            RemovedInWagtail70Warning,
        )
        return self.title

    def get_filtered_queryset(self):
        return self.filter_queryset(self.get_queryset())

    def decorate_paginated_queryset(self, object_list):
        # A hook point to allow rewriting the object list after pagination has been applied
        return object_list

    def get(self, request, *args, **kwargs):
        self.object_list = self.get_filtered_queryset()
        context = self.get_context_data()
        # Decorate the queryset *after* Django's BaseListView has returned a paginated/reduced
        # list of objects
        context["object_list"] = self.decorate_paginated_queryset(
            context["object_list"]
        )
        return self.render_to_response(context)

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["title"] = self.get_page_title()
        return context

    def render_to_response(self, context, **response_kwargs):
        if self.is_export:
            return self.as_spreadsheet(
                context["object_list"], self.request.GET.get("export")
            )
        return super().render_to_response(context, **response_kwargs)


class PageReportView(ReportView):
    results_template_name = "wagtailadmin/reports/base_page_report_results.html"
    export_headings = {
        "latest_revision_created_at": _("Updated"),
        "status_string": _("Status"),
        "content_type.model_class._meta.verbose_name.title": _("Type"),
    }
    list_export = [
        "title",
        "latest_revision_created_at",
        "status_string",
        "content_type.model_class._meta.verbose_name.title",
    ]
    context_object_name = "pages"
    permission_policy = page_permission_policy



from django.contrib.auth.models import Permission
from django.http import HttpRequest
from django.test import TestCase
from django.urls import reverse
from django.utils.http import urlencode

from wagtail import blocks
from wagtail.test.utils import WagtailTestUtils


class TestStreamFieldBlockPreviewView(WagtailTestUtils, TestCase):
    def get(self, block):
        return self.client.get(
            reverse("wagtailadmin_block_preview"),
            {"id": block.definition_prefix},
        )

    def setUp(self):
        self.user = self.login()

    def test_simple(self):
        block = blocks.CharBlock(
            label="Single-line text",
            description="A single line of text",
            preview_value="Hello, world!",
        )
        block.set_name("single_line_text")
        response = self.get(block)
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)

        html = soup.select_one("html")
        self.assertIsNotNone(html)
        self.assertEqual(html["lang"], "en")
        self.assertEqual(html["dir"], "ltr")

        robots = soup.select_one("meta[name=robots]")
        self.assertIsNotNone(robots)
        self.assertEqual(robots["content"], "noindex")

        title = soup.select_one("title")
        self.assertIsNotNone(title)
        self.assertEqual(title.text.strip(), "Preview for Single-line text (CharBlock)")

        main = soup.select_one("main")
        self.assertIsNotNone(main)
        self.assertEqual(main.text.strip(), "Hello, world!")

        wrapper = main.select_one("div.block-single_line_text")
        self.assertIsNotNone(wrapper)

    def test_nonexisting_block(self):
        response = self.client.get(reverse("wagtailadmin_block_preview"))
        self.assertEqual(response.status_code, 404)

        response = self.client.get(
            reverse("wagtailadmin_block_preview"),
            {"id": "nonexisting"},
        )
        self.assertEqual(response.status_code, 404)

    def test_no_admin_permission(self):
        self.user.is_superuser = False
        self.user.save()

        block = blocks.CharBlock()
        response = self.get(block)
        self.assertRedirects(
            response,
            reverse("wagtailadmin_login")
            + "?"
            + urlencode({"next": response.wsgi_request.get_full_path()}),
        )

    def test_minimal_permission(self):
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin",
                codename="access_admin",
            )
        )
        self.user.save()

        block = blocks.CharBlock(preview_value="Hello, world!")
        response = self.get(block)
        self.assertEqual(response.status_code, 200)

    def test_no_preview_value_no_default(self):
        block = blocks.Block()
        response = self.get(block)
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)
        main = soup.select_one("main")
        self.assertIsNotNone(main)
        self.assertEqual(main.text.strip(), "None")

    def test_preview_value_falls_back_to_default(self):
        block = blocks.IntegerBlock(default=42)
        response = self.get(block)
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)
        main = soup.select_one("main")
        self.assertIsNotNone(main)
        self.assertEqual(main.text.strip(), "42")

    def test_preview_template(self):
        class PreviewTemplateViaMeta(blocks.Block):
            class Meta:
                preview_template = "tests/custom_block_preview.html"

        class PreviewTemplateViaMethod(blocks.Block):
            def get_preview_template(self, value=None, context=None):
                return "tests/custom_block_preview.html"

        cases = [
            ("meta", PreviewTemplateViaMeta()),
            ("method", PreviewTemplateViaMethod()),
            ("kwarg", blocks.Block(preview_template="tests/custom_block_preview.html")),
        ]

        for via, block in cases:
            with self.subTest(via=via):
                response = self.get(block)
                self.assertEqual(response.status_code, 200)
                self.assertTemplateUsed(response, "tests/custom_block_preview.html")

                response = self.get(block)
                self.assertEqual(response.status_code, 200)
                self.assertTemplateUsed(response, "tests/custom_block_preview.html")

                soup = self.get_soup(response.content)
                custom_wrapper = soup.select_one("main .my-preview-wrapper")
                self.assertIsNotNone(custom_wrapper)

                custom_css = soup.select_one("link[rel=stylesheet]")
                self.assertIsNotNone(custom_css)
                self.assertEqual(custom_css["href"], "/static/css/custom.css")

                custom_js = soup.select_one("script[src]")
                self.assertIsNotNone(custom_js)
                self.assertEqual(custom_js["src"], "/static/js/custom.js")

    def test_preview_value(self):
        class PreviewValueViaMeta(blocks.Block):
            class Meta:
                preview_value = "Hello, world!"

        class PreviewValueViaMethod(blocks.Block):
            def get_preview_value(self):
                return "Hello, world!"

        cases = [
            ("meta", PreviewValueViaMeta()),
            ("method", PreviewValueViaMethod()),
            ("kwarg", blocks.Block(preview_value="Hello, world!")),
        ]

        for via, block in cases:
            with self.subTest(via=via):
                response = self.get(block)
                self.assertEqual(response.status_code, 200)
                soup = self.get_soup(response.content)
                main = soup.select_one("main")
                self.assertIsNotNone(main)
                self.assertEqual(main.text.strip(), "Hello, world!")

    def test_custom_preview_context(self):
        preview_value = "With a custom context"
        label = "Fancy block"

        class CustomPreviewContextBlock(blocks.Block):
            def get_preview_context(block, value, parent_context=None):
                self.assertEqual(value, preview_value)
                self.assertIsNotNone(parent_context)
                self.assertIsInstance(parent_context.get("request"), HttpRequest)
                self.assertIs(parent_context.get("block_def"), block)
                self.assertIs(
                    parent_context.get("block_class"),
                    CustomPreviewContextBlock,
                )
                self.assertIsInstance(
                    parent_context.get("bound_block"),
                    blocks.BoundBlock,
                )
                self.assertEqual(
                    parent_context.get("page_title"),
                    "Preview for Fancy block (CustomPreviewContextBlock)",
                )
                return {
                    **parent_context,
                    "extra": "Added by get_preview_context",
                    "page_title": "Custom title",
                }

        block = CustomPreviewContextBlock(label=label, preview_value=preview_value)
        response = self.get(block)
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)
        main = soup.select_one("main")
        self.assertIsNotNone(main)
        self.assertEqual(main.text.strip(), preview_value)
        # Ensure custom context can add and override values
        self.assertEqual(response.context["extra"], "Added by get_preview_context")
        self.assertEqual(response.context["page_title"], "Custom title")
        # Ensure that the default context is still present
        self.assertIs(response.context["block_def"], block)

    def test_static_image_preview(self):
        class StaticImagePreviewBlock(blocks.Block):
            def get_preview_context(self, value, parent_context=None):
                return {
                    "image_path": "block_previews/preview.jpg",
                    "image_description": "A preview of the block",
                }

            class Meta:
                preview_template = "tests/static_block_preview.html"

        block = StaticImagePreviewBlock()
        response = self.get(block)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "tests/static_block_preview.html")
        soup = self.get_soup(response.content)
        img = soup.select_one("html body main img")
        self.assertIsNotNone(img)
        self.assertEqual(img["src"], "/static/block_previews/preview.jpg")
        self.assertEqual(img["alt"], "A preview of the block")
